////////////////////////////////////////////////////////////////////////////
//	File:		Delegate0.h
//
//	Author:		Fabien Lebourg
//	E-mail:		fabien.lebourg@eole-soft.com
//
//	Implementation of Delegates with no argument in C++.
//
//	This code may be used in compiled form in any way you desire. This
//	file may be redistributed unmodified by any means PROVIDING it is
//	not sold for profit without the authors written consent, and
//	providing that this notice and the authors name and all copyright
//	notices remains intact.
//
//	An email letting me know how you are using it would be nice as well.
//
//	This file is provided "as is" with no expressed or implied warranty.
//	The author accepts no liability for any damage/loss of business that
//	this product may cause.


#if !defined(_DELEGATE_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_)
#define _DELEGATE_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_

// Disable long templated symbole names alert in DebugMode !
#pragma warning( disable : 4786 )

#include <assert.h>
#include <vector>
#include <list>
#include <algorithm>
using namespace std;


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


//*************************************************************************************
// For particular case (see doc)
//*************************************************************************************
#define MEMBER_FUNCTION_CALL0( CLASS ) void(CLASS::* )( )
#define MEMBER_FUNCTION_CALL_TYPEDEF0( CLASS, NAME ) void(CLASS::* NAME)( )
//*************************************************************************************


//*************************************************************************************
// CDelegateImp0_Call
//*************************************************************************************
class CDelegateImp0_Call
{
public:
   virtual void operator( ) ( ) = 0;
   virtual bool Match( int pfn, void* pobj ) const = 0;
};
//*************************************************************************************


//*************************************************************************************
// class CDelegateImp0_Call_Static
//*************************************************************************************
class CDelegateImp0_Call_Static : public CDelegateImp0_Call
{
public:
   typedef void (*PFN)( );
   CDelegateImp0_Call_Static( PFN pfn ) : m_pfn( pfn ) 
   { 
      assert( pfn ); 
   }
   CDelegateImp0_Call_Static( const CDelegateImp0_Call_Static& other ) : m_pfn( other.m_pfn )
   {
      assert( m_pfn );
   }
   virtual void operator( ) ( ) 
   { 
      (*m_pfn)( ); 
   }
   virtual bool Match( int pfn, void* pobj ) const
   {
      return (!pobj) && (m_pfn == (PFN) pfn);
   }
protected:
   PFN   m_pfn;
};
//*************************************************************************************


//*************************************************************************************
//class CDelegateImp0_Call_Member
//*************************************************************************************
template<class CLASS>
class CDelegateImp0_Call_Member : public CDelegateImp0_Call
{
public:
   typedef void (CLASS::* PFN)( );

   CDelegateImp0_Call_Member( CLASS& object, PFN pfn, int iFncID ) 
   :  m_object( object ), 
      m_pfn( pfn ), 
      m_iFncID( iFncID )
   { 
      assert( pfn ); 
   }
   CDelegateImp0_Call_Member( const CDelegateImp0_Call_Member& other ) 
   :  m_object( other.m_object ), 
      m_pfn( other.m_pfn), 
      m_iFncID( other.m_iFncID )
   { 
      assert( pfn ); 
   }
   virtual void operator( ) ( ) 
   { 
      (m_object.*m_pfn)( ); 
   }
   virtual bool Match( int iFncID, void* pobj ) const
   {
      return (&m_object == (CLASS*)pobj) && iFncID == m_iFncID;
   }
protected:
   PFN      m_pfn;
   CLASS&   m_object;
   int      m_iFncID;
};
//*************************************************************************************


//*************************************************************************************
// class CClassInfo0
//*************************************************************************************
//
// Associate an ID and a member function address
//
template<class CLASS>
class CClassInfo0
{
public:
   typedef  void (CLASS::* PFN)( );
   
   static int GetFunctionID( PFN pfn )
   {
      vector<PFN>::iterator it = find( m_vecFns.begin( ), m_vecFns.end( ), pfn );
      if( it == m_vecFns.end( ) )
      {
         m_vecFns.push_back( pfn );
         return m_vecFns.size( ) - 1;
      }
      else
      {
         return m_vecFns.begin( ) - it;
      }
   };

protected:

   static vector<PFN> m_vecFns;

};
//*************************************************************************************
template<class CLASS>
vector< CClassInfo0<CLASS>::PFN > CClassInfo0<CLASS>::m_vecFns;
//*************************************************************************************


//*************************************************************************************
// class CDelegateImp0
//*************************************************************************************
// 
// Delegate with no argument
//
template<class CS>
class CDelegateImp0  
{
public:

// --- typedefs

   typedef CDelegateImp0_Call       CALL;
   typedef list< CALL* >            CALLS;
   typedef CALLS::const_iterator    CALLS_CIT;
   typedef CALLS::iterator          CALLS_IT;

   
// --- Construction & Destruction

   CDelegateImp0( ) 
   { 
   }

   CDelegateImp0( const CDelegateImp0& other ) 
   :   m_calls( other.m_calls )
   { 
      
   }
   
   virtual ~CDelegateImp0( ) 
   { 
      Clear( ); 
   }


// --- Operations

	//-----------------------------------------------------------------
   // Execution
	//-----------------------------------------------------------------
   void  operator( )( ) const
   {
      m_cs.Lock( );
      
      CCallExecute exec;
      for_each( m_calls.begin( ), m_calls.end( ), exec );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------
   // Assign
	//-----------------------------------------------------------------
   const CDelegateImp0& operator = ( const CDelegateImp0& other )
   {
      m_cs.Lock( );
      m_calls = other.m_calls;
      m_cs.Unlock( );
      return *this;
   }
   //-----------------------------------------------------------------
  

   //-----------------------------------------------------------------   
   // Add member function address
	//-----------------------------------------------------------------   
   template<class OBJECT, class CLASS>
   void  Add( const OBJECT& object, void (CLASS::* fnc)( ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo0<CLASS>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder);
      if( it != m_calls.end( ) )
         return;

      CDelegateImp0_Call* pcall = new CDelegateImp0_Call_Member<CLASS>( *(OBJECT*)&object, fnc, iFnc );
      m_calls.push_back( pcall );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Add static function address
	//-----------------------------------------------------------------
   void  Add( void (*fnc)( ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder);
      if( it != m_calls.end( ) )
         return;

      CDelegateImp0_Call* pcall = new CDelegateImp0_Call_Static( fnc );
      m_calls.push_back( pcall );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------

   
	//-----------------------------------------------------------------   
   // Remove static function address
	//-----------------------------------------------------------------
   void  Remove( void (*fnc)( ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      if( it != m_calls.end( ) )
      {
         delete *it;
         m_calls.erase( it );
      }

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Remove member function address
	//-----------------------------------------------------------------
   template<class OBJECT, class CLASS>
   void  Remove( const OBJECT& object, void (CLASS::* fnc)( ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo0<CLASS>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) &object );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      if( it != m_calls.end( ) )
         m_calls.erase( it );

      m_cs.Unlock( );
   }      
	//-----------------------------------------------------------------


   //-----------------------------------------------------------------   
   // Has static function address ?
	//-----------------------------------------------------------------
   bool Has( void (*fnc)( ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      bool bret = (it != m_calls.end( ));

      m_cs.Unlock( );
      return bret;
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Has member function address ?
	//-----------------------------------------------------------------
   template<class OBJECT, class CLASS>
   bool Has( const OBJECT& object, void (CLASS::* fnc)( ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo0<CLASS>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) &object );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      bool bret = (it != m_calls.end( ));

      m_cs.Unlock( );
      return bret;
   }      
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Clear Delegate (remove all references)
	//-----------------------------------------------------------------
   void  Clear( )
   {
      m_cs.Lock( );

      if( m_calls.size( ) > 0 )
      {
         for_each( m_calls.begin( ), m_calls.end( ), DeleteCall );
         m_calls.erase( m_calls.begin( ), m_calls.end( ) );
      }

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------

protected:

// --- Data Members

   CALLS       m_calls;
   mutable CS  m_cs;

// --- Algo Func

	//-----------------------------------------------------------------   
   struct CCallFinder
   {
      CCallFinder( int pfn, void* pobj ) : m_pfn( pfn ), m_pobj( pobj ) { }
      int operator( )( CDelegateImp0_Call* pcall )
      {
         return pcall->Match( m_pfn, m_pobj );
      }
   protected:
      int     m_pfn;
      void*    m_pobj;
   };
	//-----------------------------------------------------------------
   struct CCallExecute
   {
      CCallExecute( ) { }
      void operator( )( CDelegateImp0_Call* pcall )
      {
         assert( pcall );
         (*pcall)( );
      }
   };
	//-----------------------------------------------------------------
   static void DeleteCall( CALL* pCall )
   {
      assert( pCall );
      delete pCall;
   }
	//-----------------------------------------------------------------   

};
//*************************************************************************************


#endif // !defined(_DELEGATE_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_)
