////////////////////////////////////////////////////////////////////////////
//	File:		Delegate1.h
//
//	Author:		Fabien Lebourg
//	E-mail:		fabien.lebourg@eole-soft.com
//
//	Implementation of Delegates with one argument in C++.
//
//	This code may be used in compiled form in any way you desire. This
//	file may be redistributed unmodified by any means PROVIDING it is
//	not sold for profit without the authors written consent, and
//	providing that this notice and the authors name and all copyright
//	notices remains intact.
//
//	An email letting me know how you are using it would be nice as well.
//
//	This file is provided "as is" with no expressed or implied warranty.
//	The author accepts no liability for any damage/loss of business that
//	this product may cause.

#if !defined(_DELEGATE1_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_)
#define _DELEGATE1_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_

// Disable long templated symbole names alert in DebugMode !
#pragma warning( disable : 4786 )

#include <assert.h>
#include <vector>
#include <list>
#include <algorithm>
using namespace std;


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


//*************************************************************************************
// For particular case (see doc)
//*************************************************************************************
#define MEMBER_FUNCTION_CALL1( CLASS, ARG ) void(CLASS::* )(ARG)
#define MEMBER_FUNCTION_CALL_TYPEDEF1( CLASS, ARG, NAME ) void(CLASS::* NAME)(ARG)
//*************************************************************************************


//*************************************************************************************
// CDelegateImp1_Call
//*************************************************************************************
template<class ARG>
class CDelegateImp1_Call
{
public:
   virtual void operator( ) ( ARG& ) = 0;
   virtual bool Match( int pfn, void* pobj ) const = 0;
};
//*************************************************************************************


//*************************************************************************************
// class CDelegateImp1_Call_Static
//*************************************************************************************
template<class ARG>
class CDelegateImp1_Call_Static : public CDelegateImp1_Call<ARG>
{
public:
   typedef void (*PFN)( ARG );
   CDelegateImp1_Call_Static( PFN pfn ) : m_pfn( pfn ) 
   { 
      assert( pfn ); 
   }
   virtual void operator( ) ( ARG& arg ) 
   { 
      (*m_pfn)( arg ); 
   }
   virtual bool Match( int pfn, void* pobj ) const
   {
      return (!pobj) && (m_pfn == (PFN) pfn);
   }
protected:
   PFN   m_pfn;
};
//*************************************************************************************


//*************************************************************************************
//class CDelegateImp1_Call_Member
//*************************************************************************************
template<class CLASS, class ARG>
class CDelegateImp1_Call_Member : public CDelegateImp1_Call<ARG>
{
public:
   typedef void (CLASS::* PFN)( ARG );

   CDelegateImp1_Call_Member( CLASS& object, PFN pfn, int iFncID ) 
   :  m_object( object ), 
      m_pfn( pfn ), 
      m_iFncID( iFncID )
   { 
      assert( pfn ); 
   }
   virtual void operator( ) ( ARG& arg ) 
   { 
      (m_object.*m_pfn)( arg ); 
   }
   virtual bool Match( int iFncID, void* pobj ) const
   {
      return (&m_object == (CLASS*)pobj) && iFncID == m_iFncID;
   }
protected:
   PFN      m_pfn;
   CLASS&   m_object;
   int      m_iFncID;
};
//*************************************************************************************


//*************************************************************************************
// class CClassInfo1
//*************************************************************************************
//
// Associate an ID and a member function address
//
template<class CLASS, class ARG>
class CClassInfo1
{
public:
   typedef  void (CLASS::* PFN)( ARG );
   
   static int GetFunctionID( PFN pfn )
   {
      vector<PFN>::iterator it = find( m_vecFns.begin( ), m_vecFns.end( ), pfn );
      if( it == m_vecFns.end( ) )
      {
         m_vecFns.push_back( pfn );
         return m_vecFns.size( ) - 1;
      }
      else
      {
         return m_vecFns.begin( ) - it;
      }
   };

protected:

   static vector<PFN> m_vecFns;

};
//*************************************************************************************
template<class CLASS, class ARG>
vector< CClassInfo1<CLASS, ARG>::PFN > CClassInfo1<CLASS, ARG>::m_vecFns;
//*************************************************************************************


//*************************************************************************************
// class CDelegateImp1
//*************************************************************************************
// 
// Delegate with one argument
//
template<class CS, class ARG>
class CDelegateImp1  
{
public:

// --- typedefs

   typedef CDelegateImp1_Call< ARG >   CALL;
   typedef list< CALL* >               CALLS;
   typedef CALLS::const_iterator       CALLS_CIT;
   typedef CALLS::iterator             CALLS_IT;
   
      
// --- Construction & Destruction

   CDelegateImp1( ) 
   { 
   }
   
   virtual ~CDelegateImp1( ) 
   { 
      Clear( ); 
   }


// --- Operations

	//-----------------------------------------------------------------
   // Execution
	//-----------------------------------------------------------------
   void  operator( ) ( ARG arg ) const
   {
      m_cs.Lock( );
      
      CCallExecute<ARG> exec( arg );
      for_each( m_calls.begin( ), m_calls.end( ), exec );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


   //-----------------------------------------------------------------
   // Assign
	//-----------------------------------------------------------------
   const CDelegateImp1& operator = ( const CDelegateImp1& other )
   {
      m_cs.Lock( );
      m_calls = other.m_calls;
      m_cs.Unlock( );
      return *this;
   }
   //-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Add member function address
	//-----------------------------------------------------------------   
   template<class OBJECT, class CLASS>
   void  Add( const OBJECT& object, void (CLASS::* fnc)( ARG ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo1<CLASS, ARG>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder);
      if( it != m_calls.end( ) )
         return;

      CDelegateImp1_Call<ARG>* pcall = new CDelegateImp1_Call_Member<CLASS, ARG>( *(OBJECT*)&object, fnc, iFnc );
      m_calls.push_back( pcall );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------
   

   //-----------------------------------------------------------------   
   // Add static function address
	//-----------------------------------------------------------------
   void  Add( void (*fnc)( ARG ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder);
      if( it != m_calls.end( ) )
         return;

      CDelegateImp1_Call<ARG>* pcall = new CDelegateImp1_Call_Static<ARG>( fnc );
      m_calls.push_back( pcall );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------

   
	//-----------------------------------------------------------------   
   // Remove static function address
	//-----------------------------------------------------------------
   void  Remove( void (*fnc)( ARG ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      if( it != m_calls.end( ) )
      {
         delete *it;
         m_calls.erase( it );
      }

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Remove member function address
	//-----------------------------------------------------------------
   template<class OBJECT, class CLASS>
   void  Remove( const OBJECT& object, void (CLASS::* fnc)( ARG ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo1<CLASS, ARG>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) &object );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      if( it != m_calls.end( ) )
         m_calls.erase( it );

      m_cs.Unlock( );
   }      
	//-----------------------------------------------------------------


   //-----------------------------------------------------------------   
   // Has static function address ?
	//-----------------------------------------------------------------
   bool Has( void (*fnc)( ARG ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      bool bret = (it != m_calls.end( ));

      m_cs.Unlock( );
      return bret;
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Has member function address ?
	//-----------------------------------------------------------------
   template<class OBJECT, class CLASS>
   bool Has( const OBJECT& object, void (CLASS::* fnc)( ARG ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo0<CLASS>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) &object );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      bool bret = (it != m_calls.end( ));

      m_cs.Unlock( );
      return bret;
   }      
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Clear Delegate (remove all references)
	//-----------------------------------------------------------------
   void  Clear( )
   {
      m_cs.Lock( );

      for_each( m_calls.begin( ), m_calls.end( ), DeleteCall );
      m_calls.erase( m_calls.begin( ), m_calls.end( ) );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------

protected:

// --- Data Members

   CALLS       m_calls;
   mutable CS  m_cs;

// --- Algo Func

	//-----------------------------------------------------------------   
   struct CCallFinder
   {
      CCallFinder( int pfn, void* pobj ) : m_pfn( pfn ), m_pobj( pobj ) { }
      int operator( )( CDelegateImp1_Call < ARG >* pcall )
      {
         return pcall->Match( m_pfn, m_pobj );
      }
   protected:
      int     m_pfn;
      void*    m_pobj;
   };
	//-----------------------------------------------------------------
   template<class ARG>
   struct CCallExecute
   {
      CCallExecute( ARG& arg ) : m_arg( arg ) { }
      void operator( )( CDelegateImp1_Call < ARG >* pcall )
      {
         assert( pcall );
         (*pcall)( m_arg );
      }
   protected:
      ARG& m_arg;
   };
	//-----------------------------------------------------------------
   static void DeleteCall( CALL* pCall )
   {
      assert( pCall );
      delete pCall;
   }
	//-----------------------------------------------------------------   

};
//*************************************************************************************


#endif // !defined(_DELEGATE1_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_)
