#ifndef _FOO_H_
#define _FOO_H_


//*************************************************************************************
// Wheel of time
//*************************************************************************************
class CClock
{
public:

// Operations

   static void  Start( ) 
   { 
      m_bRunning   = true;
      m_bTerminate = false;
      _beginthread( AsTimeGoesBy, 0, NULL );
   }

   static void  Stop( ) 
   { 
      m_bTerminate = true;
   }

   static bool  IsRunning( ) 
   {
      return m_bRunning;
   }

   static void  Format( CString& str )
   {
      const int iHours   = m_iTime / 10;
      const int iMinutes = ((m_iTime % 10) * 60) / 10;
      str.Format( "%i hour(s) and %i minute(s)", iHours, iMinutes ); 
   }


// Attributes

   static CMTDelegate1<int>&  Suscribers  ( )  { return m_delegate; }


protected:

// Members

   static int                 m_iTime;    // unit is 6 minutes
   static CMTDelegate1<int>   m_delegate;
   static CCriticalSection    m_cs;
   static bool                m_bTerminate;
   static bool                m_bRunning;


// Helpers

   static void  Advance( ) 
   { 
      m_cs.Lock( );
      ++m_iTime; 
      m_iTime = m_iTime % 240; 
      m_cs.Unlock( );
   }

   static void  AsTimeGoesBy( void* ) 
   { 
      while( !m_bTerminate )
      {
         Advance( ); 
         TRACE( "Temps coul: %d\n", m_iTime );

         if( !m_iTime )
         {
            m_bTerminate = true;
         }

         m_delegate( m_iTime ); 
         Sleep( 20L );
         
      }
      m_bRunning = false;
      _endthread( );
   }
};
//*************************************************************************************
int CClock::m_iTime( 0 );
bool CClock::m_bTerminate( false );
bool CClock::m_bRunning( false );
CMTDelegate1<int> CClock::m_delegate;
CCriticalSection  CClock::m_cs;
//*************************************************************************************



//*************************************************************************************
// Employee
//*************************************************************************************
class CEmployee
{
public:

   CEmployee( const char* psName ) : m_sName( psName ) { }
   void  Wakeup         ( )  { Notify( "woke up" ); }
   void  WelcomeOffice  ( )  { Notify( "arrived at office. " ); }
   void  LeaveOffice    ( )  { Notify( "leaved office..." ); }
   void  WelcomeHome    ( )  { Notify( "Back home! " ); }
   void  CoffeeTime     ( )  { Notify( "Had a coffee pause. " ); }
   void  LunchTime      ( )  { Notify( "ate something" ); }
   void  InBed          ( )  { Notify( "zzz... " ); }

   CMTDelegate1<const char*>& Reporter( ) { return m_reporter; }

protected:
   CString                    m_sName;
   CMTDelegate1<const char*>  m_reporter;

   void  Notify( const char* ps )
   {
      assert( ps && *ps );
      CString sMsg;
      sMsg.Format( "%s %s", m_sName, ps );
      m_reporter( sMsg );
   }
};
//*************************************************************************************


//*************************************************************************************
// Way of life... A way of delegate use
//*************************************************************************************
class CWayOfLife
{
public:
   CWayOfLife( CEmployee& employee ) 
   :  m_pEmployee( &employee ) 
   { 
   }

   void  Action( int iHour )
   {
      CMTDelegate0* pAction;
      if( m_Actions.Lookup( iHour, pAction ) )
      {
         assert( pAction );
         (*pAction)( );
      }
   }

protected:
   CMap<int, int, CMTDelegate0*, CMTDelegate0*> m_Actions;
   CEmployee* m_pEmployee;
};
//*************************************************************************************
class CWayOfLife_Theorical : public CWayOfLife
{
public:
   CWayOfLife_Theorical( CEmployee& employee ) 
   :  CWayOfLife( employee )
   {
      m_Actions[  70 ] = new CMTDelegate0;
      m_Actions[  70 ]->Add( *m_pEmployee, CEmployee::Wakeup );
      m_Actions[  89 ] = new CMTDelegate0;
      m_Actions[  89 ]->Add( *m_pEmployee, CEmployee::WelcomeOffice );
      m_Actions[ 121 ] = new CMTDelegate0;
      m_Actions[ 121 ]->Add( *m_pEmployee, CEmployee::LunchTime );
      m_Actions[ 129 ] = new CMTDelegate0;
      m_Actions[ 129 ]->Add( *m_pEmployee, CEmployee::WelcomeOffice );
      m_Actions[ 160 ] = new CMTDelegate0;
      m_Actions[ 160 ]->Add( *m_pEmployee, CEmployee::CoffeeTime );
      m_Actions[ 181 ] = new CMTDelegate0;
      m_Actions[ 181 ]->Add( *m_pEmployee, CEmployee::LeaveOffice );
      m_Actions[ 190 ] = new CMTDelegate0;
      m_Actions[ 190 ]->Add( *m_pEmployee, CEmployee::WelcomeHome );
      m_Actions[ 219 ] = new CMTDelegate0;
      m_Actions[ 219 ]->Add( *m_pEmployee, CEmployee::InBed );
   }
};
//*************************************************************************************
class CWayOfLife_Concrete : public CWayOfLife
{
public:
   CWayOfLife_Concrete( CEmployee& employee ) 
   :  CWayOfLife( employee ) 
   {
      m_Actions[ 100 ] = new CMTDelegate0;
      m_Actions[ 100 ]->Add( *m_pEmployee, CEmployee::Wakeup );
      m_Actions[ 115 ] = new CMTDelegate0;
      m_Actions[ 115 ]->Add( *m_pEmployee, CEmployee::WelcomeOffice );
      m_Actions[ 119 ] = new CMTDelegate0;
      m_Actions[ 119 ]->Add( *m_pEmployee, CEmployee::LunchTime );
      m_Actions[ 159 ] = new CMTDelegate0;
      m_Actions[ 159 ]->Add( *m_pEmployee, CEmployee::WelcomeOffice );
      m_Actions[ 160 ] = new CMTDelegate0;
      m_Actions[ 160 ]->Add( *m_pEmployee, CEmployee::CoffeeTime );
      m_Actions[ 170 ] = new CMTDelegate0;
      m_Actions[ 170 ]->Add( *m_pEmployee, CEmployee::LeaveOffice );
      m_Actions[ 220 ] = new CMTDelegate0;
      m_Actions[ 220 ]->Add( *m_pEmployee, CEmployee::WelcomeHome );
      m_Actions[ 239 ] = new CMTDelegate0;
      m_Actions[ 239 ]->Add( *m_pEmployee, CEmployee::InBed );
   }
};
//*************************************************************************************


//*************************************************************************************
// Report : Basic Callback function implementation
//*************************************************************************************
void  Report ( const char* pstr )
{
   assert( pstr && *pstr );
   cout << pstr << endl;
}
//*************************************************************************************


//*************************************************************************************
// TimeReport
//*************************************************************************************
void TimeReport ( int )
{
   CString sTime;
   CClock::Format( sTime );
   cout << (const char*) sTime << endl;
}
//*************************************************************************************


#endif _FOO_H_