// DimConverter.h: interface for the DimConverter class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DimConverter_H__6C99DAC1_74D6_4D90_A749_AEA53811FE61__INCLUDED_)
#define AFX_DimConverter_H__6C99DAC1_74D6_4D90_A749_AEA53811FE61__INCLUDED_


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//*************************************************************************************
// class DimConverter  
//*************************************************************************************
//
// Calculation of Linear to N-dimensions index conversions.
// 
template<int NBDIM>
class DimConverter  
{

public:

// --- Construction & Destruction

   // '...' represents the dimensions of the 
   // Logical N-Dim of the linear Container
   // (one integer for each dimension: NBDIM integers)
   DimConverter( int ilast, ... )
   {
      va_list args;
      va_start( args, ilast );
      
      int I = 0;
      m_Coefs[ I ]   = 1;
      m_Coefs[ ++I ] = ilast;
      while( ++I < NBDIM )
      {
         int idim = va_arg( args, int );
         m_Coefs[ I ] = idim * m_Coefs[ I - 1 ];
      }
      va_end( args );
   }

	~DimConverter( )
   {
   }


// --- Operations

   // Set logical dimensions
   // (one integer for each dimension: NBDIM integers)
   void SetDim( int ilast, ... )
   {
      va_list args;
      va_start( args, ilast );
      
      int I = 0;
      m_Coefs[ I ]   = 1;
      m_Coefs[ ++I ] = ilast;
      while( ++I < NBDIM )
      {
         int idim = va_arg( args, int );
         m_Coefs[ I ] = idim * m_Coefs[ I - 1 ];
      }
      va_end( args );
   }

   // '...' is a list of integers (one integer for each dimension: NBDIM-1 integers )
   int ToIdx ( int ilast, ... ) const
   {
      va_list args;
      va_start( args, ilast );
      
      int I = 0;
      int ivaldim = va_arg( args, int );
      int index = ilast;
      
      while( ++I < NBDIM )
      {
         index += ivaldim * m_Coefs[ I ];
      }
      va_end( args );
      return index;
   }

   // '...' is a list of integers references (one integer for each dimension: NBDIM integers )
   void ToDim ( int idx, ... ) const
   {
      int values[ NBDIM ];
      int iTmp = idx;

      for( int I = NBDIM; --I >=0; )
      {
         values[ I ] = iTmp / m_Coefs[ I ];
         iTmp = iTmp % m_Coefs[ I ];
      }

      va_list args;
      va_start( args, idx );

      I = NBDIM;
      while( --I >= 0 )
      {
         int& ivaldim = va_arg( args, int& );
         ivaldim = values[ I ];
      }
      va_end( args );
   }

protected:

// --- Data Members

   int  m_Coefs[ NBDIM ];

};
//*************************************************************************************


#endif // !defined(AFX_DimConverter_H__6C99DAC1_74D6_4D90_A749_AEA53811FE61__INCLUDED_)
